import typer
import sys

add_select = "À quel projet ajouter une tâche ?"
create_name = "Quel est le nom de ce nouveau projet ?"
delete_select_all = "Quel projet voulez-vous supprimer ?"
delete_select = "Dans quel projet se trouve la tâche à supprimer ?"
description = "Comment décrire le projet ?"
done_select = "À quel projet appartient la tâche à terminer ?"
done_rank = "Quel est le rang de la tâche à terminer ?"
due_date = "Quelle est la date butoir pour ce projet (JJ/MM/AAA HH:MM:SS) ?"
info_select = "Entrer le Numéro ou nom du projet à consulter | (C)réer un projet"
modify_rank = "Quel est le rang de la tâche à modifier ?"
modify_select = "Quel projet est à modifier ?"
modify_task_select = "À quel projet appartient la tâche à modifier ?"
move_rank = "Quel est le rang de la tâche à déplacer ?"
move_select = "À quel projet appartient la tâche à déplacer ?"
task_body = "En quoi consiste cette tâche ?"
relocate_new = "Quel est le nouvel identifiant du projet"
relocate_select = "Quel projet est à deplacer ?"
rename_name = "Quel est le nouveau nom de ce projet ?"
rename_select = "Quel projet est à renommer ?"
start_rank = "Quel est le rang de la tâche à commencer ?"
start_select = "À quel projet appartient la tâche à commencer ?"
undo_rank = "Quel est le rang de la tâche à retravailler ?"
undo_select = "À quel projet appartient la tâche à retravailler ?"
unstart_rank = "Quel est le rang de la tâche à arrêter ?"
unstart_select = "À quel projet appartient la tâche à arrêter ?"


quit_button = " | (Q)uitter "


def input(prompt):
    var = typer.prompt(prompt + quit_button)
    low = var.lower().strip()
    if low == "q":
        print("OK Bye !")
        sys.exit(0)
    return var.strip()



from colorama import Fore, Style
from datetime import datetime


def labelize(str):
    label = str.rjust(16) + " : "
    return dim(label)


def justify(str):
    justified = str.rjust(16) + " - "
    return justified


def margin(str):
    return " ".rjust(13) + str


def projectNum(id):
    return f"{id}".zfill(3)


def taskNum(id):
    return f"{id}".zfill(2)


def dim(str):
    dimmed = Style.DIM + str + Style.RESET_ALL
    return dimmed


def green(str):
    colored = Fore.GREEN + str + Style.RESET_ALL
    return colored


def yellow(str):
    colored = Fore.YELLOW + str + Style.RESET_ALL
    return colored


def blue(str):
    colored = Fore.BLUE + str + Style.RESET_ALL
    return colored


def red(str):
    colored = Fore.RED + str + Style.RESET_ALL
    return colored


def success(str):
    msg = green("\nOK!\n") + str + "\n"
    print(msg)


def warning(str):
    msg = yellow("\nATTENTION !\n") + str
    return msg


def safe(str):
    msg = blue("\nPAS DE PANIQUE !\n") + str + "\n"
    print(msg)


def error(e):
    str = f"{e}"
    msg = red("\nOupsy...\n" + str + "\n")
    print(msg)


def toFrenchDate(mysqlDate):
    if mysqlDate:
        return mysqlDate.strftime("%d/%m/%Y %H:%M:%S")
    return None


def toUSDate(frenchDate):
    return datetime.strptime(frenchDate, "%d/%m/%Y %H:%M:%S")


def show_status(obj, long: bool = False):
    if long:
        status = "À faire"
        if obj.started:
            status = yellow("En cours  ")
        if obj.completed:
            status = green("Terminé  ")
    else:
        status = " "
        if obj.started:
            status = yellow(" ")
        if obj.completed:
            status = green(" ")
    return status


def main_header():
    text = "Projets"

    return "\n" + dim(margin(text)) + "\n"


def project_header(project):
    id = projectNum(project.ID)
    name = project.name
    description = project.description
    if not description:
        description = dim("Ajouter une description 'lefaire description [projectID]'")
    header = justify(id) + name
    subHeader = labelize("Description") + description
    return "\n\n" + header + "\n" + subHeader


def projects(projectList):
    todo = []
    done = []
    for project in projectList:
        box = show_status(project)
        id = projectNum(project.ID)
        name = project.name
        line = justify(id) + box + " " + name
        if project.completed:
            done.append(line)
        else:
            todo.append(line)

    if todo:
        todo.append("\n")
    if done:
        done.append("\n")

    view = "\n".join(todo) + "\n".join(done)
    return view


def list(projectList):
    view = main_header()
    if projectList:
        view += projects(projectList)
    else:
        view += dim(
            "Vous n'avez pas encore de projets.\nAjoutez-en avec la commande `create`"
        )
    print(view)


def tasks(taskList, long=False):
    todo = []
    done = []
    if not taskList:
        placeholder = (
            "\n"
            + labelize("Tâches")
            + dim(
                "Ce projet n'a encore aucune tâche. Pour ajouter un tâche 'lefaire add [projectID]'\n"
            )
        )
        return placeholder
    for task in taskList:
        box = show_status(task)
        rank = taskNum(task.rank)
        body = task.body
        if task.completed:
            line = dim(justify(rank)) + box + " " + dim(body)
            done.append(line)
        else:
            line = justify(rank) + dim(box) + " " + body
            todo.append(line)

    otherDone = ""
    if not long and len(done) > 7:
        otherDone = f"{len(done)-7} tâches effectuées +\n"
        done = done[slice(len(done) - 7, len(done))]
    if todo:
        todo.append("\n")
    if done:
        done.append("\n")

    view = (
        "\n" + "\n".join(todo) + " ".rjust(14) + dim(otherDone) + "\n" + "\n".join(done)
    )
    return view


def project_info(project, taskList, long=False):
    header = project_header(project) + "\n"
    status = labelize("Statut") + show_status(project, True) + "\n"
    created = labelize("Crée le") + toFrenchDate(project.created) + "\n"
    if project.started:
        started = labelize("Commencé le") + toFrenchDate(project.started) + "\n"
    else:
        started = labelize("Commencé le") + dim(
            "Pour commencer un projet, commencer une de ses tâches 'lefaire start [projectID] [rank]'\n"
        )

    if project.completed:
        completed = labelize("Terminé le") + toFrenchDate(project.completed) + "\n"
        completionTime = labelize("Durée totale") + str(
            project.completed - project.started
        )
        completion = completed + completionTime + "\n"
    else:
        completion = labelize("Terminé le") + dim(
            "Pour compléter un projet, compléter toutes ses tâches 'lefaire done [projectID] [rank]'\n"
        )
    if project.dueDate:
        dueDate = labelize("Date butoir") + toFrenchDate(project.dueDate) + "\n"
    else:
        dueDate = labelize("Date butoir") + dim(
            "Ajouter une date butoir 'lefaire due [projectID]'\n"
        )
    subtasks = tasks(taskList, long)
    infos = header + status + created + dueDate + started + completion + subtasks
    print(infos)


def task_info(task):
    rank = taskNum(task.rank)
    header = justify(rank) + task.body + "\n"
    status = labelize("Statut") + show_status(task, True) + "\n"
    created = labelize("Crée le") + toFrenchDate(task.created) + "\n"
    if task.started:
        started = labelize("Commencé le") + toFrenchDate(task.started) + "\n"
    else:
        started = labelize("Commencé le") + dim(
            "Commencer une tâche 'lefaire start [taskID] [rank]'\n"
        )
    if task.completed:
        completed = labelize("Terminé le") + toFrenchDate(task.completed) + "\n"
        completionTime = labelize("Durée totale") + str(task.completed - task.started)
        completion = completed + completionTime + "\n"
    else:
        completion = labelize("Terminé le") + dim(
            "Compléter une tâche 'lefaire done [projectID] [rank]"
        )
    infos = "\n" + header + status + created + started + completion + "\n"
    print(infos)


def create(project):
    return success(
        "Le projet "
        + projectNum(project.ID)
        + " - "
        + green(project.name)
        + " a bien été créé !"
    )


def rename(nb, name):
    return success(
        "Le projet " + projectNum(nb) + f" a bien été renommé '{green(name)}'."
    )


def relocate(project):
    return success(
        f"Le projet {green(project.name)} porte désormais le numéro {green(projectNum(project.ID))}."
    )


def desc(project):
    return success(
        f"La description du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été modifiée."
    )


def due(project):
    return success(
        f"La date butoir du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été modifiée."
    )


def add(task, project):
    return success(
        f"La tâche {green(task.body)} a bien été ajoutée au projet '{green(project.name)}'."
    )


def edit(task, project):
    return success(
        f"La tâche n°{task.rank} du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été renommée '{green(task.body)}'."
    )


def start(task, project):
    return success(
        f"La tâche '{green(task.body)}' du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été commencée."
    )


def unstart(task, project):
    return success(
        f"La tâche '{green(task.body)}' du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été arrêtée."
    )


def complete(task, project):
    return success(
        f"La tâche '{green(task.body)}' du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été terminée."
    )


def commit_cmd(task):
    return f"\n git commit -m '{task.body}'"


def commit_prompt(cmd):
    return warning(
        f"On est dans le {yellow('bon dossier')} ?\nOn peut exécuter la commande : {yellow(cmd)} ? "
    )


def commit_success():
    return success(f"La commande a bien été {green('exécutée')}")


def uncomplete(task, project):
    return success(
        f"La tâche '{green(task.body)}' du projet {projectNum(project.ID)} - '{green(project.name)}' n'est pas terminée."
    )


def move(task, project, dir):
    if dir.lower() == "up":
        dir = "vers le haut"
    elif dir.lower() == "down":
        dir = "vers le bas"
    elif dir.lower() == "top":
        dir = "tout en haut"
    else:
        dir = "tout en bas"

    return success(
        f"La tâche '{green(task.body)}' du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été déplacée {dir}."
    )


def delete_task(task, project):
    return success(
        f"La tâche '{green(task.body)}' du projet {projectNum(project.ID)} - '{green(project.name)}' a bien été supprimée."
    )


def delete_project(project):
    return success(
        f"Le projet {projectNum(project.ID)} - '{green(project.name)}' a bien été supprimé."
    )


def delete_all_prompt():
    return warning(f"Voulez-vous supprimer un projet en {yellow('ENTIER')} ?")


def delete_task_warning(task, project):
    return warning(
        f"La tâche '{yellow(task.body)}' du projet '{yellow(project.name)}' est sur le point d'être supprimée.\n\nContinuer ? "
    )


def delete_project_warning(project):
    return warning(
        f"Le projet  '{yellow(project.name)}' est sur le point d'être supprimé.\n\nContinuer ? "
    )


def delete_task_safe(task, project):
    return safe(
        f"La tâche '{blue(task.body)}' du projet '{blue(project.name)}' est conservée."
    )


def delete_project_safe(project):
    return safe(f"Le projet  '{blue(project.name)}' est conservé.")



from dataclasses import dataclass
from lefaire.connect import connect


@dataclass
class Project:
    ID: int
    name: str
    created: str
    started: str
    completed: str
    dueDate: str
    description: str


@dataclass
class Task:
    ID: int
    projectID: int
    rank: int
    body: str
    created: str
    started: str
    completed: str


# Retrieve all projects
def get_projects():
    conn = connect()
    cur = conn.cursor()
    projects = []
    cur.execute(
        "SELECT ID, Name, Created, Started, CompletionDate, DueDate, Description from Project"
    )

    for (
        ID,
        name,
        created,
        started,
        completed,
        dueDate,
        description,
    ) in cur:
        project = Project(ID, name, created, started, completed, dueDate, description)
        projects.append(project)

    conn.close()
    return projects


# Retrieve on particular project
def get_single_project(projectID):

    conn = connect()
    cur = conn.cursor()

    if not projectIDExists(cur, projectID):
        raise Exception("Projet introuvable")
    project = None
    cur.execute(
        "SELECT Name, Created, Started, CompletionDate, DueDate, Description from Project WHERE ID = ?",
        (projectID,),
    )
    for (
        name,
        created,
        started,
        completed,
        dueDate,
        description,
    ) in cur:
        project = Project(
            projectID,
            name,
            created,
            started,
            completed,
            dueDate,
            description,
        )

    conn.close()
    return project


# Show all tasks for a project
def get_tasks(projectID: int):

    conn = connect()
    cur = conn.cursor()

    if not (projectIDExists(cur, projectID)):
        raise Exception("Projet introuvable")
    tasks = []
    cur.execute(
        "SELECT ID, Rank, Body, Created, Started, CompletionDate FROM Task WHERE ProjectID = ? ORDER BY Rank;",
        (projectID,),
    )

    for ID, rank, body, created, started, completed in cur:
        task = Task(ID, projectID, rank, body, created, started, completed)
        tasks.append(task)

    tasks = clean_ranks(cur, tasks)
    conn.commit()
    conn.close()
    return tasks


def clean_ranks(cur, tasksList):

    # split todo and done()
    todo = []
    done = []
    for task in tasksList:
        if task.completed:
            done.append(task)
        else:
            todo.append(task)
    newList = done + todo
    newOrder(cur, newList)
    return newList


# get a specific task from a project
def get_single_task(projectID, rank):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    if not taskRankExists(cur, projectID, rank):
        raise NameError("Tâche introuvable")

    cur.execute(
        "SELECT ID, Body, Created, Started, CompletionDate FROM Task WHERE ProjectID = ? AND Rank = ?;",
        (projectID, rank),
    )
    for ID, body, created, started, completed in cur:
        task = Task(ID, projectID, rank, body, created, started, completed)

    conn.close()
    return task


def get_task_ID(projectID, rank):
    conn = connect()
    cur = conn.cursor()

    if not taskRankExists(cur, projectID, rank):
        raise NameError("Tâche introuvable")

    cur.execute(
        "SELECT ID FROM Task WHERE ProjectID = ? AND Rank = ?;",
        (projectID, rank),
    )
    id = cur.fetchone()[0]
    conn.close()
    return id


def get_task_by_ID(taskID):

    conn = connect()
    cur = conn.cursor()

    task = None
    cur.execute(
        "SELECT ProjectID, Rank, Body, Created, Started, CompletionDate FROM Task WHERE ID = ?;",
        (taskID,),
    )
    for projectID, rank, body, created, started, completed in cur:
        task = Task(taskID, projectID, rank, body, created, started, completed)
    conn.close()
    return task


def projectNameExists(cur, projectName: str):
    cur.execute(
        "SELECT EXISTS (SELECT * FROM Project WHERE Name = ?) ;",
        (projectName,),
    )
    bool = cur.fetchone()[0]
    return bool


def projectIDExists(cur, projectID: int):
    cur.execute(
        "SELECT EXISTS (SELECT * FROM Project WHERE ID = ?) ;",
        (projectID,),
    )
    bool = cur.fetchone()[0]
    return bool


def toID(input: str):

    conn = connect()
    cur = conn.cursor()

    cur.execute("SELECT ID FROM Project WHERE Name = ?;", (input,))
    res = cur.fetchone()
    conn.close()
    if res is not None:
        id = res[0]
        return id
    else:
        try:
            return int(input)
        except Exception:
            raise Exception(f"'{input}' est introuvable")
            pass


# Add a project
def create_project(projectName="Nouveau Projet", description=None):

    conn = connect()
    cur = conn.cursor()

    projectName = projectName.strip()
    if description:
        description = description.strip()
    if projectNameExists(cur, projectName):
        raise NameError("Le projet existe déjà !")

    if not description:
        description = None

    cur.execute(
        "INSERT INTO Project (Name, Created, Description) VALUES ( ?, CURRENT_TIMESTAMP, ?);",
        (projectName, description),
    )

    conn.commit()
    conn.close()


def update_rank(cur, ID, newRank):
    cur.execute("UPDATE Task SET Rank = ? WHERE ID = ?;", (newRank, ID))


def newOrder(cur, list):

    for i in range(len(list)):
        task = list[i]
        update_rank(cur, task.ID, i + 1)


def taskBodyExists(cur, projectID, body):
    cur.execute(
        "SELECT EXISTS (SELECT * FROM Task WHERE ProjectID = ? AND Body = ?);",
        (projectID, body),
    )
    bool = cur.fetchone()[0]
    return bool


def taskRankExists(cur, projectID, rank):
    cur.execute(
        "SELECT EXISTS (SELECT * FROM Task WHERE ProjectID = ? AND Rank = ?);",
        (projectID, rank),
    )
    bool = cur.fetchone()[0]
    return bool


# Add a task to a project
def add_task(projectID: int, body: str):

    project = get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    if taskBodyExists(cur, projectID, body):
        raise Exception("La tâche existe déjà")

    body = body.strip()
    tasksList = get_tasks(projectID)
    rank = len(tasksList) + 1
    cur.execute(
        "INSERT INTO Task (ProjectID, Rank, Body) VALUES (?, ?, ?);",
        (projectID, rank, body),
    )

    conn.commit()
    conn.close()

    if project.completed:
        uncomplete_project(projectID)


# Update project
def rename_project(projectID, newName):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    newName = newName.strip()
    cur.execute("UPDATE Project SET Name = ? WHERE ID = ?", (newName, projectID))

    conn.commit()
    conn.close()


def update_description(projectID, newDescription):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    newDescription = newDescription.strip()
    cur.execute(
        "UPDATE Project SET Description = ? WHERE ID = ?", (newDescription, projectID)
    )
    conn.commit()
    conn.close()


def update_due_date(projectID, dueDate):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    cur.execute("UPDATE Project SET DueDate = ? WHERE ID = ?", (dueDate, projectID))
    conn.commit()
    conn.close()


def start_project(projectID):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    cur.execute(
        "UPDATE Project SET Started = CURRENT_TIMESTAMP WHERE ID = ?", (projectID,)
    )
    conn.commit()
    conn.close()


def unstart_project(projectID):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    cur.execute("UPDATE Project SET Started = NULL WHERE ID = ?", (projectID,))
    conn.commit()
    conn.close()


def complete_project(projectID):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    cur.execute(
        "UPDATE Project SET CompletionDate = CURRENT_TIMESTAMP  WHERE ID = ?",
        (projectID,),
    )
    conn.commit()
    conn.close()


def uncomplete_project(projectID):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    cur.execute(
        "UPDATE Project SET CompletionDate = NULL WHERE ID = ?",
        (projectID,),
    )
    conn.commit()
    conn.close()


def relocate_project(projectID, newID):

    get_single_project(projectID)

    conn = connect()
    cur = conn.cursor()

    if projectIDExists(cur, newID):
        raise Exception(f"Le projet n°{newID} existe déjà !")

    cur.execute(
        "UPDATE Project SET ID = ? WHERE ID = ?",
        (
            newID,
            projectID,
        ),
    )
    # changer le project ID de ces taches
    cur.execute(
        "UPDATE Task SET ProjectID = ? WHERE ProjectID = ?",
        (
            newID,
            projectID,
        ),
    )

    conn.commit()
    conn.close()
    change_auto_increment()


def change_auto_increment():

    conn = connect()
    cur = conn.cursor()

    try:
        # find the bigger ID
        cur.execute("SELECT MAX(ID) FROM Project")
        maxID = cur.fetchone()[0]
        # Make the auto-increment the next number
        if maxID:
            auto_increment = f"{maxID + 1}"
            cur.execute(f"ALTER TABLE Project AUTO_INCREMENT={auto_increment};")
            cur.execute("COMMIT")
    except Exception:
        raise Exception("Erreur de manipulation des auto-increments ")
    finally:
        conn.commit()
        conn.close()


# Update a task
def update_body(projectID, rank, newBody):

    get_single_task(projectID, rank)

    conn = connect()
    cur = conn.cursor()

    newBody = newBody.strip()
    cur.execute(
        "UPDATE Task SET Body = ? WHERE ProjectID = ? AND Rank = ?",
        (newBody, projectID, rank),
    )
    conn.commit()
    conn.close()


def swap_rank(projectID, currentRank, newRank):
    firstTask = get_single_task(projectID, currentRank)
    secondTask = get_single_task(projectID, newRank)
    conn = connect()
    cur = conn.cursor()
    cur.execute("UPDATE Task SET Rank = ? WHERE ID=?", (newRank, firstTask.ID))
    cur.execute("UPDATE Task SET Rank = ? WHERE ID=?", (currentRank, secondTask.ID))
    conn.commit()
    conn.close()


def move_up(projectID, rank):
    task = get_single_task(projectID, rank)
    if rank == 1:
        raise Exception(
            f"'{task.body}' est la première tâche, elle ne peut donc pas monter plus haut."
        )
    swap_rank(projectID, rank, rank - 1)


def move_down(projectID, rank):

    task = get_single_task(projectID, rank)
    try:
        get_single_task(projectID, rank + 1)
    except Exception:
        raise Exception(
            f"'{task.body}' est la dernière tâche, elle ne peut donc pas descendre plus bas."
        )
    swap_rank(projectID, rank, rank + 1)


def move_ext(projectID, rank, ext):

    task = get_single_task(projectID, rank)
    ext = ext.lower()
    tasksList = get_tasks(projectID)

    conn = connect()
    cur = conn.cursor()
    # split todo and done
    todo = []
    done = []
    if rank == 1 and ext == "top":
        raise Exception(
            f"'{task.body}' est déjà la première tâche, elle ne peut donc pas monter plus haut."
        )
    if rank == len(tasksList) and ext == "bottom":
        raise Exception(
            f"'{task.body}' est déjà la dernière tâche, elle ne peut donc pas descendre plus bas."
        )
    for current in tasksList:
        if current != task:
            if current.completed:
                done.append(current)
            else:
                todo.append(current)
    if ext == "top":
        if task.completed:
            done.insert(0, task)
        else:
            todo.insert(0, task)
    else:
        if task.completed:
            done.append(task)
        else:
            todo.append(task)

    newList = done + todo
    newOrder(cur, newList)
    conn.commit()
    conn.close()


def move(projectID, rank, dir):

    dir = dir.lower()
    if dir == "up":
        move_up(projectID, rank)
    elif dir == "down":
        move_down(projectID, rank)
    elif dir == "top" or dir == "bottom":
        move_ext(projectID, rank, dir)
    else:
        raise Exception("La direction ne peut être que 'top', 'bottom', 'up' ou 'down'")


def start_task(projectID, rank):

    task = get_single_task(projectID, rank)

    conn = connect()
    cur = conn.cursor()

    if task.started:
        raise Exception(f"La tâche '{task.body}' est déjà commencée.")
    if task.completed:
        raise Exception(f"La tâche '{task.body}' est déjà terminée.")
    cur.execute(
        "UPDATE Task SET started = CURRENT_TIMESTAMP WHERE ProjectID = ? AND Rank = ?",
        (projectID, rank),
    )
    conn.commit()
    conn.close()
    project = get_single_project(projectID)
    if not project.started:
        start_project(projectID)
    if task.completed:
        uncomplete_task(projectID, rank)


def noTasksStarted(tasksList):
    for task in tasksList:
        if task.started or task.completed:
            return False
    return True


def checkifStarted(cur, projectID):
    tasksList = get_tasks(projectID)
    if noTasksStarted(tasksList):
        unstart_project(projectID)


def unstart_task(projectID, rank):

    task = get_single_task(projectID, rank)
    conn = connect()
    cur = conn.cursor()

    if not task.started:
        raise Exception(f"La tâche '{task.body}' n'est pas encore commencée.")
    if task.completed:
        raise Exception(f"La tâche '{task.body}' est déjà terminée.")
    cur.execute(
        "UPDATE Task SET started = NULL WHERE ProjectID = ? AND Rank = ?",
        (projectID, rank),
    )
    conn.commit()
    conn.close()
    project = get_single_project(projectID)
    if project.started:
        checkifStarted(cur, projectID)
    updatedTask = get_single_task(projectID, rank)
    if updatedTask.completed:
        uncomplete_task(projectID, rank)


def allTasksComplete(tasksList):
    for task in tasksList:
        if not task.completed:
            return False
    return True


def checkCompleteness(cur, projectID):
    tasksList = get_tasks(projectID)
    if not tasksList:
        uncomplete_project(projectID)
        unstart_project(projectID)
    elif allTasksComplete(tasksList):
        complete_project(projectID)
    else:
        uncomplete_project(projectID)


def complete_task(projectID, rank):

    task = get_single_task(projectID, rank)

    if task.completed:
        raise Exception(f"La tâche '{task.body}' est déjà terminée")

    if int(rank) > 1:
        move_ext(projectID, rank, "top")
    newTask = get_task_by_ID(task.ID)
    rank = newTask.rank
    conn = connect()
    cur = conn.cursor()
    if not task.started:
        start_task(projectID, rank)
    cur.execute(
        "UPDATE Task SET CompletionDate = CURRENT_TIMESTAMP  WHERE ID = ? ",
        (task.ID,),
    )
    conn.commit()
    conn.close()
    checkCompleteness(cur, projectID)
    get_tasks(projectID)


def uncomplete_task(projectID, rank):

    task = get_single_task(projectID, rank)

    conn = connect()
    cur = conn.cursor()

    if not task.completed:
        raise Exception(f"La tâche '{task.body}' n'est pas terminée")
    cur.execute(
        "UPDATE Task SET CompletionDate = NULL WHERE ProjectID = ? AND Rank = ?",
        (projectID, rank),
    )
    conn.commit()
    conn.close()
    checkCompleteness(cur, projectID)


def delete_project(projectID):

    conn = connect()
    cur = conn.cursor()

    if not projectIDExists(cur, projectID):
        raise Exception("Le project n'existe pas ")
    cur.execute("DELETE FROM Project WHERE ID = ?", (projectID,))
    cur.execute("DELETE FROM Task WHERE ProjectID = ?;", (projectID,))
    conn.commit()
    conn.close()
    change_auto_increment()


def delete_task(taskID):
    task = get_task_by_ID(taskID)

    conn = connect()
    cur = conn.cursor()

    projectID = task.projectID
    cur.execute("DELETE FROM Task WHERE ID = ?;", (taskID,))
    conn.commit()
    conn.close()
    checkifStarted(cur, projectID)
    checkCompleteness(cur, projectID)



import mariadb
import secrets
import string
from getpass import getpass
from dotenv import load_dotenv, set_key
from pathlib import Path
from lefaire.view import error, success, warning
import sys
import os


def get_env_file_path():
    config_dir = os.path.join(str(Path.home()), ".config", "lefaire")
    os.makedirs(config_dir, exist_ok=True)
    return os.path.join(config_dir, ".env")


env_path = get_env_file_path()


def clean_db(cur):
    cur.execute(
        """
        DELETE FROM Task; 
        """
    )
    cur.execute(
        """
        DELETE FROM Project; 
        """
    )
    increment_reset(cur)


def setup():
    try:
        ## Ask for a root access
        (root_password, root_host) = get_credentials()

        # Establish a root connexion
        root_conn = mariadb.connect(user="root", host=root_host, password=root_password)
        success("Connexion établie !")
        cur = root_conn.cursor()

        ## delete user if it exists
        flush_user(cur, root_host)

        # create databases
        create_databases(cur)

        # generate securely a password
        new_password = generate_password()

        ## Create new user with secure password and give access to databases
        grant_privileges(cur, new_password, root_host)

        # Store the password to the env
        set_key(env_path, "LEFAIRE_PASSWORD", new_password)
        set_key(env_path, "LEFAIRE_HOST", root_host)

        # load variables from .env

        load_dotenv(env_path)
        root_conn.commit()
        root_conn.close()
        return

    except Exception:
        raise Exception("Accès refusé ...")


def get_credentials():
    print("Connexion à MariaDB")
    print("Utilisateur : root")
    host = input("Host (laissez vide pour localhost) :")
    if not host:
        host = "localhost"

    root_password = getpass("Mot de passe : ")
    return (root_password, host)


def flush_user(cur, host):
    print(
        "Supression de l'utilisateur MariaDB (anciennement MySQL) 'lefaire' s'il existe... "
    )
    cur.execute(f"DROP USER IF EXISTS 'lefaire'@'{host}'")


def create_tables(cur, db):
    cur.execute(
        f" CREATE TABLE IF NOT EXISTS {db}.Project ( `ID` int(11) NOT NULL AUTO_INCREMENT, `Created` timestamp NOT NULL,`Started` timestamp NULL DEFAULT NULL,`CompletionDate` timestamp NULL DEFAULT NULL,`DueDate` datetime DEFAULT NULL,`Description` text DEFAULT NULL,`Name` varchar(255) NOT NULL,PRIMARY KEY (`ID`)) ENGINE=InnoDB AUTO_INCREMENT=1 DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci "
    )
    cur.execute(
        f"""
                CREATE TABLE IF NOT EXISTS {db}.Task (
                `ID` int(11) NOT NULL AUTO_INCREMENT,
                `ProjectID` int(11) NOT NULL,
                `Rank` int(11) NOT NULL,
                `Body` text NOT NULL,
                `Created` timestamp NULL DEFAULT current_timestamp(),
                `Started` timestamp NULL DEFAULT NULL,
                `CompletionDate` timestamp NULL DEFAULT NULL,
                PRIMARY KEY (`ID`)
)
            """
    )


def create_databases(cur):
    print("Création de la base de donnée ...")
    cur.execute("CREATE DATABASE IF NOT EXISTS lefaire")
    create_tables(cur, "lefaire")  # if doesn't exist
    print("Création de la base de donnée de test ...")
    cur.execute("CREATE DATABASE IF NOT EXISTS lefaire_test")
    create_tables(cur, "lefaire_test")  # if doesn't exist


def generate_password():
    print("Création d'un mot de passe sécurisé ...")
    safe_punctuation = "!@#$%/&*()[]{}_-+=:;.,<>?"
    alphabet = string.ascii_letters + string.digits + safe_punctuation
    password = "".join(secrets.choice(alphabet) for i in range(16))
    return password


def grant_privileges(cur, password, host):
    print("Autorisation d'accès à la base de données lefaire...")
    cur.execute(
        f"""
            GRANT ALL PRIVILEGES ON lefaire.* TO 'lefaire'@'{host}' IDENTIFIED BY '{password}' ;
        """
    )
    print("Autorisation d'accès à la base de données de test...")
    cur.execute(
        f"GRANT ALL PRIVILEGES ON lefaire_test.* TO 'lefaire'@'{host}' IDENTIFIED BY '{password}' ;"
    )


def connect(test: bool = False):

    load_dotenv(env_path)
    DB_PASSWORD = os.environ.get("LEFAIRE_PASSWORD")

    if not DB_PASSWORD:
        setup(env_path)
        DB_PASSWORD = os.environ.get("LEFAIRE_PASSWORD")

    DB = os.environ.get("LEFAIRE_DB")

    if not DB:
        set_key(env_path, "LEFAIRE_DB", "lefaire")
        DB = "lefaire"

    try:
        conn = get_connect(DB)
        return conn

    except mariadb.Error as e:
        pass
        error(f"Erreur de connexion ({e})")
        setup()
        load_dotenv(env_path)
        print(warning("La connexion a été réinitialisée. Veuillez réessayer"))
        sys.exit(1)


def get_connect(db):
    conn = mariadb.connect(
        user="lefaire",
        host=os.environ["LEFAIRE_HOST"],
        password=os.environ["LEFAIRE_PASSWORD"],
        database=db,
    )
    return conn


def increment_reset(cur):
    cur.execute("ALTER TABLE Project AUTO_INCREMENT=1")
    cur.execute("ALTER TABLE Task AUTO_INCREMENT=1")



import typer
import os
import lefaire.view as view
import lefaire.data as data
import lefaire.prompt as prompt
from lefaire.connect import setup

from .prompt import input
from typing_extensions import Annotated

app = typer.Typer()
move_app = typer.Typer()
app.add_typer(move_app, name="move", help="Déplacer une tâche dans la liste de tâche.")


@app.command()
def add(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    body: Annotated[str, typer.Argument(help="Corps de la tâche")] = "",
    top: Annotated[
        bool,
        typer.Option(
            "-t",
            "--top",
            help="Mettre la tâche tout en haut ?",
        ),
    ] = False,
):
    """
    Ajouter une tâche.
    """

    try:
        if not select:
            list()
            select = input(prompt.add_select)

        projectID = data.toID(select)

        if not body:
            list(projectID)
            body = input(prompt.task_body)

        project = data.get_single_project(projectID)
        data.add_task(projectID, body)
        taskList = data.get_tasks(projectID)
        task = taskList[-1]

        if top:
            data.move_ext(projectID, task.rank, "top")
            task = data.get_task_by_ID(task.ID)

        list(projectID)
        view.add(task, project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def create(
    name: Annotated[str, typer.Argument(help="Nom du projet")] = "",
    description: Annotated[str, typer.Argument(help="Description du projet")] = "",
):
    """
    Créer un projet.
    """
    try:
        if not name:
            name = input(prompt.create_name)

        data.create_project(name, description)
        project = data.get_projects()[-1]
        list()
        view.create(project)

    except NameError as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def delete(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
    all: Annotated[bool, typer.Option("-a", "--all", "-p", "--project")] = False,
):
    """
    Supprimer un projet. Pour supprimer une tâche, ajouter son rang
    """
    confirm = False
    try:
        if not select:
            if not all:
                all = typer.confirm(view.delete_all_prompt())
                confirm = True
            if all:
                select = input(prompt.delete_select_all)
            else:
                select = input(prompt.delete_select)

        projectID = data.toID(select)

        project = data.get_single_project(projectID)
        if not rank:
            if not all and not confirm:
                confirm = typer.confirm(view.delete_all_prompt())
                all = confirm
            if all:
                delete_project(project)
            else:
                list(project.ID)
                rank = typer.prompt("Quelle tâche doit être supprimée ? ")
        if rank:
            task = data.get_single_task(project.ID, rank)

            delete_task(project, task)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


def delete_project(project):

    confirm = typer.confirm(view.delete_project_warning(project))
    if confirm:
        data.delete_project(project.ID)
        list()
        view.delete_project(project)
    else:
        list()
        view.delete_project_safe(project)


def delete_task(project, task):

    confirm = typer.confirm(view.delete_task_warning(task, project))
    if confirm:
        data.delete_task(task.ID)
        list(project.ID)
        view.delete_task(task, project)
    else:
        list(project.ID)
        view.delete_task_safe(task, project)


@app.command()
def desc(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    description: Annotated[str, typer.Argument(help="Description du projet")] = "",
):
    """
    Ajouter ou modifier la description d'un projet.
    """

    try:
        if not select:
            list()
            select = input(prompt.modify_select)

        projectID = data.toID(select)

        if not description:
            list(projectID)
            description = input(prompt.description)

        data.update_description(projectID, description)
        project = data.get_single_project(projectID)
        list(projectID)
        view.desc(project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def done(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
    commit: Annotated[bool, typer.Option("-g", "-c", "--git", "--commit")] = False,
):
    """
    Marquer une tâche comme terminée.
    """
    try:
        if not select:
            list()
            select = input(prompt.done_select)

        projectID = data.toID(select)

        if not rank:
            list(projectID)
            rank = input(prompt.done_rank)

        project = data.get_single_project(projectID)
        task = data.get_single_task(projectID, rank)

        data.complete_task(projectID, rank)
        task = data.get_task_by_ID(task.ID)
        list(projectID, task.rank)
        view.complete(task, project)

        # run git command if wanted
        cmd = view.commit_cmd(task)
        if commit:
            sure = True
        else:
            sure = typer.confirm(view.commit_prompt(cmd))
        if sure:
            os.system(cmd)
            list(projectID)
            view.commit_success()

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def due(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    date: Annotated[str, typer.Argument(help="Date butoir du projet")] = "",
):
    """
    Ajouter ou modifier une date butoir (JJ/MM/AAA HH:MM:SS).
    """

    try:
        if not select:
            list()
            select = input(prompt.modify_select)

        projectID = data.toID(select)

        if not date:
            list(projectID)
            date = input(prompt.due_date)

        date = view.toUSDate(date)
        data.get_single_project(projectID)
        data.update_due_date(projectID, date)
        project = data.get_single_project(projectID)
        list(projectID)
        view.due(project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def edit(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
    body: Annotated[str, typer.Argument(help="Corps de la tâche")] = "",
):
    """
    Modifier le corps d'une tâche.
    """
    try:
        if not select:
            list()
            select = input(prompt.modify_task_select)

        projectID = data.toID(select)

        if not rank:
            list(projectID)
            rank = input(prompt.modify_rank)
            rank = int(rank)

        if not body:
            list(projectID, rank)
            body = input(prompt.task_body)

        project = data.get_single_project(projectID)
        data.update_body(projectID, rank, body)
        task = data.get_single_task(projectID, rank)
        list(projectID, rank)
        view.edit(task, project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


def list(select: int = 0, rank: int = 0, long: bool = False):
    """
    Lister les projets et leurs tâches.
    """
    if rank:
        task = data.get_single_task(select, rank)
        view.task_info(task)

    elif select:
        projectID = data.toID(select)
        project = data.get_single_project(projectID)
        taskList = data.get_tasks(projectID)
        view.project_info(project, taskList, long)

    else:
        projectsList = data.get_projects()
        view.list(projectsList)


def project_menu(select):
    choice = typer.prompt(
        "(A)jouter une tâche | (I)nfo d'une tâche | (R)enommer | Changer la (D)escription | Changer la date (B)utoir | (S)upprimer | (Q)uitter "
    )

    choice = choice.lower().strip()
    if choice == "a":
        add(select)
    elif choice == "i":
        rank = typer.prompt("Quel est le rang de la tâche à consulter ? ")
        info(select, rank)
    elif choice == "r":
        rename(select)
    elif choice == "d":
        desc(select)
    elif choice == "b":
        due(select)
    elif choice == "s":
        delete(select, "", True)
    elif choice == "q":
        print("\n")
    else:
        try:

            int(choice)
            data.get_single_task(select, choice)
        except Exception as e:
            pass
            view.error(e)
        else:
            info(select, choice)


def task_menu(select, rank):
    rank = int(rank)
    task = data.get_single_task(select, rank)
    menu = "(E)diter | "
    if not task.started:
        menu += "(C)ommencer | "
    if task.started and not task.completed:
        menu += "(A)rrêter | "
    if not task.completed:
        menu += "(T)erminer | "
    if task.completed:
        menu += "(R)etravailler | "
    menu += "(D)éplacer | (S)upprimer | (Q)uitter "

    choice = typer.prompt(menu)
    choice = choice.lower().strip()
    if choice == "e":
        edit(select, rank)
    if choice == "c":
        start(select, rank)
    if choice == "a":
        unstart(select, rank)
    if choice == "t":
        done(select, rank)
    if choice == "r":
        undo(select, rank)
    if choice == "s":
        delete(select, rank)
    if choice == "q":
        info(select)
    if choice == "d":
        dir = typer.prompt(
            "Déplacer vers le (H)aut | Déplacer vers le (B)as | Mettre en (P)remier | Mettre en (D)ernier "
        )
        dir = dir.lower().strip()
        if dir == "h":
            dir = "up"
        if dir == "b":
            dir = "down"
        if dir == "p":
            dir = "top"
        if dir == "d":
            dir = "bottom"
        move(dir, select, rank)


@app.command()
def info(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
    long: Annotated[
        bool,
        typer.Option(
            "-l", "-v", "--long", "--verbose", help="Affiche toutes les tâches ?"
        ),
    ] = False,
):
    """
    Détailler les infos sur un projet.
    """

    try:
        if rank:
            list(select, rank)
            task_menu(select, rank)
        else:
            if not select:
                list()
                select = input(prompt.info_select)

            if select.lower() == "c":
                create()

            else:
                projectID = data.toID(select)
                list(projectID, 0, long)
                project_menu(projectID)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def init():
    """
    Créé un mot de passe et un utilisateur pour se connecter à MariaDB
    """
    try:
        setup()

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def relocate(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    new: Annotated[int, typer.Argument(help="Nouveau Numéro ou nom du projet")] = 0,
):
    """
    Changer le Numéro ou nom d'identifiant d'un projet
    """

    try:
        if not select:
            list()
            select = input(prompt.relocate_select)
        projectID = data.toID(select)

        project = data.get_single_project(projectID)

        if not new:
            custom_prompt = prompt.relocate_new + project.name + " ?"
            new = input(custom_prompt)

        data.relocate_project(projectID, new)
        project = data.get_single_project(projectID)
        list()
        view.relocate(project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def rename(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    name: Annotated[str, typer.Argument(help="Nouveau nom du projet")] = "",
):
    """
    Renommer un projet.
    """
    try:
        if not select:
            list()
            select = input(prompt.rename_select)

        projectID = data.toID(select)

        if not name:
            name = input(prompt.rename_name)

        data.rename_project(projectID, name)
        list()
        view.rename(projectID, name)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def start(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Marquer une tâche comme commencée.
    """
    try:
        if not select:
            list()
            select = input(prompt.start_select)

        projectID = data.toID(select)

        if not rank:
            list(projectID)
            rank = input(prompt.start_rank)

        project = data.get_single_project(projectID)
        data.start_task(projectID, rank)
        task = data.get_single_task(projectID, rank)
        list(projectID)
        view.start(task, project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def undo(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Marquer une tâche comme non terminée.
    """

    try:
        if not select:
            list()
            select = input(prompt.undo_select)

        projectID = data.toID(select)

        if not rank:
            list(projectID)
            rank = input(prompt.undo_rank)

        project = data.get_single_project(projectID)
        task = data.get_single_task(projectID, rank)
        data.uncomplete_task(projectID, rank)
        task = data.get_task_by_ID(task.ID)
        list(projectID)
        view.uncomplete(task, project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@app.command()
def unstart(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Marquer une tâche comme non commencée.
    """

    try:
        if not select:
            list()
            select = input(prompt.unstart_select)

        projectID = data.toID(select)

        if not rank:
            list(projectID)
            rank = input(prompt.unstart_rank)

        project = data.get_single_project(projectID)
        data.unstart_task(projectID, rank)
        task = data.get_single_task(projectID, rank)
        list(projectID)
        view.unstart(task, project)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


def move(
    dir: Annotated[str, typer.Argument(help="Direction")],
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Déplacer une tâche dans la liste de tâche.
    """
    try:
        if not select:
            list()
            select = input(prompt.move_select)

        projectID = data.toID(select)

        if not rank:
            list(projectID)
            rank = input(prompt.move_rank)

        rank = int(rank)
        task = data.get_single_task(projectID, rank)
        data.move(projectID, rank, dir)
        project = data.get_single_project(projectID)
        list(projectID)
        view.move(task, project, dir)

    except Exception as e:
        view.error(e)
        raise typer.Exit()


@move_app.command("up")
def move_up(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Déplacer une tâche vers le haut dans la liste de tâche.
    """
    move("up", select, rank)


@move_app.command("down")
def move_down(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Déplacer une tâche vers le bas dans la liste de tâche
    """
    move("down", select, rank)


@move_app.command("top")
def move_top(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Déplacer une tâche tout en haut de la liste de tâche.
    """
    move("top", select, rank)


@move_app.command("bottom")
def move_bottom(
    select: Annotated[str, typer.Argument(help="Numéro ou nom du projet")] = "",
    rank: Annotated[int, typer.Argument(help="Rang de la tâche")] = 0,
):
    """
    Déplacer une tâche tout en vas de la liste de tâche.
    """
    move("bottom", select, rank)


if __name__ == "__main__":
    app()
